/*
 * Copyright (c) 1998 Sasha Vasko <sashav@sprintmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define DO_CLOCKING      

#define TRUE 1
#define FALSE

#include "../../configure.h"

#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/time.h>
#ifdef DO_CLOCKING      
#include <time.h>
#endif

#ifdef ISC			/* Saul */
#include <sys/bsdtypes.h>	/* Saul */
#endif /* Saul */

#include <stdlib.h>
#if defined ___AIX || defined _AIX || defined __QNX__ || defined ___AIXV3 || defined AIXV3 || defined _SEQUENT_
#include <sys/select.h>
#endif

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#ifdef I18N
#include <X11/Xlocale.h>
#endif

#include "../../include/aftersteplib.h"
#include "parser.h"

#ifdef DEBUG_PARSER
#define LOG1(a)       fprintf( stderr, a ); 
#define LOG2(a,b)    fprintf( stderr, a, b );
#define LOG3(a,b,c)    fprintf( stderr, a, b, c );
#define LOG4(a,b,c,d)    fprintf( stderr, a, b, c, d );
#else
#define LOG1(a)
#define LOG2(a,b)
#define LOG3(a,b,c)
#define LOG4(a,b,c,d)
#endif

HashValue GetHashValue( char* text, int hash_size )
{
  register int i ;
  HashValue hash_value = (HashValue)text[0];
      for( i = 1 ; i < sizeof(HashValue)-sizeof(char) ; i++ )
      {
          if( text[i] == '\0' || isspace(text[i]) ) break ;
	  hash_value |= (((HashValue)(tolower(text[i])))<<i);
      }
  
    return hash_value%hash_size ;
}

void BuildHash( SyntaxDef* syntax, TermDef** term_hash, HashValue hash_size )
{
  HashValue hash_value ;
  TermDef* current ;
  int i ;
    for( i = 0 ; i < hash_size ; i++ ) term_hash[i] = NULL ;
    for( i = 0 ; syntax->terms[i].keyword ; i++ )
    {
	hash_value = GetHashValue( syntax->terms[i].keyword, hash_size );
	if( term_hash[hash_value] )
	{
	    for( current = term_hash[hash_value] ; 
	         current->brother ; 
		 current = current->brother );
	    current->brother = &(syntax->terms[i]) ;
	}else term_hash[hash_value] = &(syntax->terms[i]);
    }
}

/* reader initialization */
ConfigDef* InitConfigReader( SyntaxDef* syntax, const char* filename, FILE* fd, HashValue hash_size, int (*special)(struct config_definition* conf_def, FreeStorageElem** storage))
{
  ConfigDef* new_conf ;

    if( filename == NULL && fd == NULL ) return NULL;  
    
    new_conf = (ConfigDef*)safemalloc( sizeof(ConfigDef) );
    new_conf->syntax=syntax ;
    new_conf->term_hash_size = hash_size ;
    new_conf->term_hash = (TermDef**)safemalloc( sizeof( TermDef*)*hash_size);
    BuildHash( syntax, new_conf->term_hash, hash_size );
    
    new_conf->special = special;
    if( filename != NULL ) new_conf->realfilename  = PutHome( filename );
    else new_conf->realfilename= NULL ;
    
    if( fd )
    {
	new_conf->fd = fd;
	new_conf->bNeedToCloseFile = 0 ;
    }else
    {
	new_conf->fd = fopen( new_conf->realfilename, "rt");
	new_conf->bNeedToCloseFile = 1 ;
    }
    
    /* allocated to store lines read from the file */
    new_conf->buffer  = (char*) safemalloc( MAXLINELENGTH );
    
    /* this is the current parsing information */
    new_conf->tline = NULL ;     
    new_conf->current_term = NULL ;
    new_conf->current_data = (char*) safemalloc( MAXLINELENGTH );
    new_conf->current_data_len = 0 ;
    new_conf->cursor = &(new_conf->buffer[0]) ;
    new_conf->buffer[0] = '\0' ;
    return new_conf ;
}

/* debugging stuff */
#ifdef DEBUG_PARSER
void PrintConfigReader(ConfigDef* config)
{
int i ;
    fprintf( stderr, "\nSentence Terminator: [0x%2.2x]", config->syntax->terminator );
    fprintf( stderr, "\nConfig Terminator:   [0x%2.2x]", config->syntax->file_terminator );
    fprintf( stderr, "\nConfig filename:     [%s]", config->realfilename );
    fprintf( stderr, "\nTerm's hash table:" );
    for(i = 0 ; i < config->term_hash_size ; i++ )
    {
      TermDef* pterm = config->term_hash[i];
    	if( pterm ) fprintf( stderr, "\n  Hash value %d:", i );
	while( pterm )
	{
	    fprintf( stderr, "\n\t\t[%lx][%d][%d][%s]", pterm->flags, pterm->type, pterm->id, pterm->keyword );
	    pterm = pterm->brother ;
	}
    }
    fprintf( stderr, "\nEnd of Term's hash table." );    
}

void PrintFreeStorage(FreeStorageElem* storage)
{
int i ;
    while( storage )
    {
	fprintf( stderr, "\nTerm's keyword: [%s]", storage->term->keyword );
	fprintf( stderr, "\nData:" );
	for(i = 0 ; i < storage->argc ; i++ )
    	    fprintf( stderr, "\n  %d:\t[%s]", i, storage->argv[i]  );
	storage=storage->next ;
    }
}
#endif

/* reader de-initialization */
void DestroyConfigReader(ConfigDef* config)
{
    if( config->realfilename ) free(config->realfilename);
    free( config->buffer );
    free( config->current_data );
    free( config->term_hash );
    if( config->bNeedToCloseFile && config->fd ) fclose( config->fd ) ;
    free( config );
}
/* this function finds next valid statement:
   - not comments, 
   - not an empty line, 
   - if prepended with * - should have MyName at the beginning
   and :
   - sets tline to the beginning of the keyword (not MyName),
   - copies arguments into the current_data
   - sets current_data_len
   it will read data from the file as needed untill the end of file 
   or config->syntax.file_terminator or EOF is reached
   Return: NULL if end of config reached, otherwise same as tline.
 */
char* GetNextStatement(ConfigDef* config)
{
  char* cur = config->cursor ; /* don't forget to save it back ! */
    while(1)
    {
	for( ; *cur!= '\0' ; cur++ )
	{
	    if( *cur == config->syntax->file_terminator ) return NULL ;
	    if( !isspace( *cur ) )	    
	    {
   	      register int i ; 

		if( *cur == COMMENTS_CHAR ) break ; /* skip entire line */
		if( *cur == MYNAME_CHAR )
		{ /* check if we have MyName here */
		    for( i = 0 ; *cur!= '\0' && MyName[i]!= '\0' ; i++ )
		    	if( tolower(MyName[i]) != tolower(*(++cur)) ) break ;
		    if( MyName[i] != '\0' ) break ; /* skip entire line */
		    cur++ ;
		}
		config->tline = cur ; /*that will be the begginnig of the term*/
		/* now we should copy everything from after the first space to
		   config->current_data and set current_data_len ; */
		for( ; !isspace( *cur ) && *cur!=config->syntax->terminator; cur++ );
		for( ; isspace( *cur ) && *cur && *cur!=config->syntax->terminator ; cur++ );
		for( i = 0 ; *(cur+i) && *(cur+i)!=config->syntax->terminator ; i++ )
		    config->current_data[i] = *(cur+i) ;
		config->current_data_len = i ;
		config->current_data[i] = '\0';
		config->cursor = cur+i ; /* Saving position for future use */
		return config->tline ;
	    }	
	}
	/* reading file here */
	if( config->fd )
	{
	    if( !fgets( config->buffer, MAXLINELENGTH, config->fd )) return NULL;
	    cur = &(config->buffer[0]);
/*fprintf( stderr, "\n line read from the file :[%s]",cur );*/
	}else return NULL ;
    }
    return NULL ;
}

/* this one will identify Term using Hash table
   it will seT :
   - current_term to the term found
   Returns : NULL if no terms found, otherwise same as current_term.
 */
TermDef* FindStatementTerm(ConfigDef* config)
{
  HashValue hash_value  ;
  TermDef* pterm ;
    
    hash_value= GetHashValue( config->tline, config->term_hash_size );
    for( pterm = config->term_hash[hash_value] ; pterm ; pterm=pterm->brother )
    	if( mystrncasecmp( config->tline, pterm->keyword, pterm->keyword_len ) == 0 )
	{
	    config->current_term = pterm ; 
	    return pterm ;	    
	}
    
    return NULL ;
}

/* main parsing procedure */
int ParseConfig( ConfigDef* config, FreeStorageElem** storage )
{
  FreeStorageElem** pNext = storage ;
    /* get line */
    while( GetNextStatement(config))
    {/* untill not end of text */
      TermDef* pterm ;
#ifdef DEBUG_PARSER
        fprintf( stderr, "\nSentence Found:[%s]\n,\tData=\t[%s]", config->tline, config->current_data  );
        fprintf( stderr, "\nLooking for the Term..." );
#endif
	/* find term */
	if( (pterm = FindStatementTerm(config)))
	{
#ifdef DEBUG_PARSER
        fprintf( stderr, "\nTerm Found:[%s]", config->current_term->keyword );
#endif
	    if( pterm->flags&TF_SPECIAL_PROCESSING )
	    {
		if( config->special ) 
		    (*(config->special))(config, storage); 
	    }else
	    {/* create freestorage elem */
		*pNext = (FreeStorageElem*)safemalloc( sizeof( FreeStorageElem ) );
		(*pNext)->term = pterm ;
#ifdef DEBUG_PARSER
                fprintf( stderr, "\ncurr_data:[%s], len = %d", config->current_data, config->current_data_len );
#endif

		if( config->current_data_len <= 0 )
		{
		    (*pNext)->argv = NULL ;
		    (*pNext)->argc = 0 ;
		}else 
		{
		  int i, count ;
		  char* cur ;

		    if( !(pterm->flags&TF_DONT_SPLIT) )		
		    {
		        cur = config->current_data ;
			for( (*pNext)->argc = 0 ; *cur ; )
			{
			    for( count = 0 ; !isspace(*cur)&& *cur ; count++ ) 
				cur++;			    
			    while( isspace(*cur) && *cur) cur++;	
			    if( count ) (*pNext)->argc++;
			}
		    }else (*pNext)->argc = 1 ;
		    
		    (*pNext)->argv = (char**) safemalloc( sizeof(char*)*((*pNext)->argc));
		    (*pNext)->argv[0] = (char*)safemalloc( config->current_data_len );
		    if( pterm->flags&TF_DONT_SPLIT )
			strcpy( (*pNext)->argv[0], config->current_data );
		    else 
		    {
			cur = config->current_data ;
			for( i = 0 ; i<(*pNext)->argc ; i++ )
			{
			    for( count = 0 ; !isspace(*cur)&& *cur ; count++ ) 
				(*pNext)->argv[i][count] = *(cur++) ;
    			    while( isspace(*cur) && *cur) cur++;	
			    if( !count ) break ;

			    (*pNext)->argv[i][count] = '\0' ;
			    if( (i+1)<(*pNext)->argc )
				(*pNext)->argv[i+1] = &((*pNext)->argv[i][count+1]);

			}
		    }
		}
		(*pNext)->next = NULL ;
		pNext = &((*pNext)->next) ;
	    }
	}
    }/* end utill */
    return 1 ;
}

/* memory deallocation */
void DestroyFreeStorage(FreeStorageElem** storage )
{
    if( storage )
	if( *storage )
	{
	    DestroyFreeStorage( &((*storage)->next) );
	    /* that will deallocate everything */
	    if( (*storage)->argv[0] ) free( (*storage)->argv[0] ); 
	    free( *storage );
	    *storage = NULL ;
	}
}

